void IO_Init()
{
   // Init I/O.
   SPI_SlaveInit();
   PWM_Init();
   ADC_Init();

   // Direction of the digial io.
   DDRA = 0x00;
   DDRC = 0xff;
}

signed char IO_TestAck(unsigned char tx, unsigned char rx)
{
   if(tx == 255){	// If the ack has overflowed.
      if(rx != 0) return -1;
   }
   else{ // Normal ack.
      if(rx != tx + 1) return -1;
   }
   // Ack was correct.
   return 0;
}

signed char IO_NewCommand(unsigned char code)
{
   // Checks the code, and start the correct command.
   // Only test against 5 MSB, since 3LSB only contain information about channel.
   switch(code & 0xf8){
      case CODE_AIN:
         IO_AnalogIn(code);
         break;
      case CODE_AOUT:
         IO_AnalogOut(code);
         break;
      case CODE_DIN:
         IO_DigitalIn(code);
         break;
      case CODE_DOUT:
         IO_DigitalOut(code);
         break;
#ifdef BUFFERING_MODE
      case CODE_AIN_EN:
         IO_AnalogInEnable(code);
         break;
#endif
      case CODE_AOUT_RES:
         IO_AnalogOutRes(code);
         break;
      default:
         //Prints cmd to UART if it didn't match any codes.
         UART_Print('e', (code & 0xf8));
         break;
   }
}

signed char IO_AnalogIn(unsigned char code){
   unsigned char rx, data[2];
   char ret;

   // 1st byte: check received DUMMY, return code + 1.
   ret = SPI_SlaveTransferByte(code + 1, &rx, TIMEOUT);
   if(ret < 0) return ret;
   if(rx != DUMMY) return -1;

#ifdef INTERRUPT_MODE
   // Starts conversion and polls until the conversion is complete.
   ADC_StartConversion(CODE_TO_CH_AIN(code));
   while(!ADC_CheckConversion(data));

   // Send interrupt to AVR32.
   set_bit(PORTD, 7);
   _delay_us(2);
   clear_bit(PORTD, 7);
#endif

#ifdef BUFFERING_MODE
   // Store data that should be sent to master.
   data[0] = dataH[CODE_TO_CH_AIN(code)];
   data[1] = dataL[CODE_TO_CH_AIN(code)];
#endif

   // 2nd byte: check received DUMMY, return high byte value.
   ret = SPI_SlaveTransferByte(data[0], &rx, TIMEOUT);
   if(ret < 0) return ret;
   if(rx != DUMMY) return -1;

   // 3rd byte: check received high byte value + 1, return low byte value.
   ret = SPI_SlaveTransferByte(data[1], &rx, TIMEOUT);
   if(ret < 0) return ret;
   ret = IO_TestAck(data[0], rx);
   if(ret < 0) return ret;

   // 4th byte: check received low byte value + 1, return DUMMY.
   ret = SPI_SlaveTransferByte(DUMMY, &rx, TIMEOUT);
   if(ret < 0) return ret;
   ret = IO_TestAck(data[1], rx);
   if(ret < 0) return ret;

   // 5th byte: check received MASTER_OK, return SLAVE_OK.
   ret = SPI_SlaveTransferByte(SLAVE_OK, &rx, TIMEOUT);
   if(ret < 0) return ret;
   if(rx != MASTER_OK) return -1;

   // Debug printing, if enabled.
   UART_Print('c', code);
   UART_Print('h', data[0]);
   UART_Print('l', data[1]);

   return 0;
}

signed char IO_AnalogOut(unsigned char code)
{
   unsigned char valueH, valueL, rx;
   char ret;

   // 1st byte: store received valueH, return code + 1.
   ret = SPI_SlaveTransferByte(code + 1, &valueH, TIMEOUT);
   if(ret < 0) return ret;

   // 2nd byte: store received valueL, return valueH + 1.
   ret = SPI_SlaveTransferByte(valueH + 1, &valueL, TIMEOUT);
   if(ret < 0) return ret;

   // 3rd byte: check received DUMMY, return valueL + 1.
   ret = SPI_SlaveTransferByte(valueL + 1, &rx, TIMEOUT);
   if(ret < 0) return ret;
   if(rx != DUMMY) return -1;

   // 4th byte: check received MASTER_OK, return SLAVE_OK.
   ret = SPI_SlaveTransferByte(SLAVE_OK, &rx, TIMEOUT);
   if(ret < 0) return ret;
   if(rx != MASTER_OK) return -1;

   // When the message is confirmed the mask and data are applied to digital out.
   PWM_SetOutValue(CODE_TO_CH_AOUT(code), CODE_TO_SCH_AOUT(code), valueH, valueL);

   // Debug printing, if enabled.
   UART_Print('c', code);
   UART_Print('h', valueH);
   UART_Print('l', valueL);
}

signed char IO_DigitalIn(unsigned char code)
{
   unsigned char rx, value;
   char ret;

   // Only read from the port once during communication
   value = PINA;

   // 1st byte: check received DUMMY, return code + 1.
   ret = SPI_SlaveTransferByte(code + 1, &rx, TIMEOUT);
   if(ret < 0) return ret;
   if(rx != DUMMY) return -1;

   // 2nd byte: check received DUMMY, return value.
   ret = SPI_SlaveTransferByte(value, &rx, TIMEOUT);
   if(ret < 0) return ret;
   if(rx != DUMMY) return -1;

   // 3rd byte: check received value + 1, return DUMMY.
   ret = SPI_SlaveTransferByte(DUMMY, &rx, TIMEOUT);
   if(ret < 0) return ret;
   ret = IO_TestAck(value, rx);
   if(ret < 0) return ret;

   // 4th byte: check received MASTER_OK, return SLAVE_OK.
   ret = SPI_SlaveTransferByte(SLAVE_OK, &rx, TIMEOUT);
   if(ret < 0) return ret;
   if(rx != MASTER_OK) return -1;

   // Debug printing, if enabled.
   UART_Print('c', code);
   UART_Print('v', value);

   return 0;
}

signed char IO_DigitalOut(unsigned char code)
{
   unsigned char mask, value, rx;
   char ret, i;

   // 1st byte: store received mask, return code + 1.
   ret = SPI_SlaveTransferByte(code + 1, &mask, TIMEOUT);
   if(ret < 0) return ret;

   // 2nd byte: store received value, return mask + 1.
   ret = SPI_SlaveTransferByte(mask + 1, &value, TIMEOUT);
   if(ret < 0) return ret;

   // 3rd byte: check received DUMMY, return value + 1.
   ret = SPI_SlaveTransferByte(value + 1, &rx, TIMEOUT);
   if(ret < 0) return ret;
   if(rx != DUMMY) return -1;

   // 4th byte: check received MASTER_OK, return SLAVE_OK.
   ret = SPI_SlaveTransferByte(SLAVE_OK, &rx, TIMEOUT);
   if(ret < 0) return ret;
   if(rx != MASTER_OK) return -1;

   // When the message is confirmed the mask and data are applied to digital out.
   for(i=0;i<8;i++){
      if(test_bit(mask, i)){
         if(test_bit(value, i)) set_bit(PORTC, i);
         else clear_bit(PORTC, i);
      }
   }

   // Debug printing, if enabled.
   UART_Print('c', code);
   UART_Print('m', mask);
   UART_Print('v', value);
}

#ifdef BUFFERING_MODE
signed char IO_AnalogInEnable(unsigned char code)
{
   unsigned char enable, rx;
   char ret;

   // 1st byte: store received enable, return code + 1.
   ret = SPI_SlaveTransferByte(code + 1, &enable, TIMEOUT);
   if(ret < 0) return ret;

   // 2nd byte: store received DUMMY, return enable + 1.
   ret = SPI_SlaveTransferByte(enable + 1, &rx, TIMEOUT);
   if(ret < 0) return ret;
   if(rx != DUMMY) return -1;

   // 3rd byte: check received MASTER_OK, return SLAVE_OK.
   ret = SPI_SlaveTransferByte(SLAVE_OK, &rx, TIMEOUT);
   if(ret < 0) return ret;
   if(rx != MASTER_OK) return -1;

   // When the message is confirmed the channel are enabled or disabled.
   if(enable == 0){
      ADC_DisableChan(CODE_TO_CH_AIN(code));
   }
   else{
      ADC_EnableChan(CODE_TO_CH_AIN(code));
   }

   // Debug printing, if enabled.
   UART_Print('c', code);
   UART_Print('e', enable);
}
#endif

signed char IO_AnalogOutRes(unsigned char code)
{
   unsigned char resolution, chEnable, rx;
   char ret;

   // 1st byte: store received resolution, return code + 1.
   ret = SPI_SlaveTransferByte(code + 1, &resolution, TIMEOUT);
   if(ret < 0) return ret;

   // 2nd byte: check received DUMMY, return resolution + 1.
   ret = SPI_SlaveTransferByte(resolution + 1, &rx, TIMEOUT);
   if(ret < 0) return ret;
   if(rx != DUMMY) return -1;

   // 3rd byte: check received MASTER_OK, return SLAVE_OK.
   ret = SPI_SlaveTransferByte(SLAVE_OK, &rx, TIMEOUT);
   if(ret < 0) return ret;
   if(rx != MASTER_OK) return -1;

   // When the message is confirmed the analog out resolution of the selected channel will be updated.
   PWM_SetResolution(CODE_TO_CH_AOUT(code), resolution);

   // Debug printing, if enabled.
   UART_Print('c', code);
   UART_Print('r', resolution);
}
