#include "avr32io.h"

static int __init avr32io_Init(void)
{
   int ret;

   if(!(device = kzalloc(sizeof(struct avr32io_device), GFP_KERNEL)))
   {
      printk(KERN_ALERT "avr32io_Init: kzalloc failed\n");
      return -ENOMEM;
   }

   // Major/minor number allocation
   ret = alloc_chrdev_region(&device->number, 0, COUNT_TOTAL, "avr32io");
   if(ret < 0){
      printk(KERN_ALERT "avr32io_Init: alloc_chrdev_region failed %i\n", ret);
      goto errRegion;
   }

   // Adding cdev.
   device->charDevice = cdev_alloc();
   device->charDevice->ops = &fops;
   device->charDevice->owner = THIS_MODULE;
   ret = cdev_add(device->charDevice, device->number, COUNT_TOTAL);
   if(ret < 0){
      printk(KERN_ALERT "avr32io_Init: charDevice failed %i\n", ret);
      goto errCharDevice;
   }

   // Registration of the driver with the linux device model
   ret = spi_register_driver(&avr32io_driver);
   if(ret < 0){
      printk(KERN_ALERT "avr32io_Init: spi_register_driver failed %i\n", ret);
      goto errRegDriver;
   }

#ifdef INTERRUPT_MODE
	// Registration of IRQ
   ret = request_irq(EIM_IRQ_BASE, irqFunk, IRQF_TRIGGER_RISING, "avr32io_interrupt", NULL);
   if(ret < 0){
      DEBUG("avrIO_Init: request_irq failed %i\n", ret);
      goto errRegIRQ;
   }
   device->irqID = ret;
#endif

   printk(KERN_INFO "avr32io_Init: sucessfull major: %i\n", MAJOR(device->number));

   //Return sucessfull
   return 0;

   //Error, roll back resourses.
#ifdef INTERRUPT_MODE
   errRegIRQ:
      spi_unregister_driver(&avr32io_driver);
#endif
   errRegDriver:
      cdev_del(device->charDevice);
   errCharDevice:
      unregister_chrdev_region(device->number, 4);
   errRegion:
      kfree(device);
   return ret;
}

static void __exit avr32io_Exit(void)
{
#ifdef INTERRUPT_MODE
   // Free the IRQ.
   free_irq(EIM_IRQ_BASE,NULL);
#endif

   // Unregister the driver from the linux device model
   spi_unregister_driver(&avr32io_driver);

   // Unregister the character device from the kernel
   cdev_del(device->charDevice);

   // Unregister major/minor numbers allocated
   unregister_chrdev_region(device->number, COUNT_TOTAL);

   // Free memory used by the device struct.
   kfree(device);

   printk(KERN_INFO "avr32io_Exit: module exit\n");
}

static int __devinit avr32io_Probe(struct spi_device * spi)
{
   int ret;

   // Initialize the spi struct
   device->spi = spi;
   spi->mode = SPI_MODE_0;
   spi->bits_per_word = 0x08; // 8-bits per transfer.
   spi->max_speed_hz = 2100000; //highest value that works is 2100000 @ 16MHz
   spi->chip_select = 0x02; // Uses slave number 2.

   // Setup the SPI driver and channel
   ret = spi_setup(spi);
   if(ret < 0){
      printk(KERN_ALERT "avr32io_Probe: spi_setup failed %d\n", ret);
      return ret;
   }

   dev_set_drvdata(&spi->dev, device);
   return 0;
}

static int avr32io_Open(struct inode *node,struct file *filp)
{
   int minor, ret;

   minor = iminor(filp->f_dentry->d_inode);
   DEBUG("avr32io_Open: minor %i\n", minor);

   // Take mutex.
   ret = down_interruptible(&spiMutex);
   if(ret < 0) return -ERESTARTSYS;

   // Check if SPI is used.
   if(device->busy == 0){
      // SPI busy, return error.
      ret = -EBUSY;
   }
   else{
      // SPI not busy, mark as busy and return success.
      device->busy == 1;
      ret = 0;
   }

   // Returns the mutex and return if the SPI was busy or not.
   up(&spiMutex);
   return ret;
}

static int avr32io_Release(struct inode *node,struct file *filp)
{
   DEBUG("avr32io_Release\n");

   // Take mutex.
   ret = down_interruptible(&spiMutex);
   if(ret < 0) return -ERESTARTSYS;

   // Mark the SPI as not busy.
   device->busy == 0;

   // Returns the mutex and returns success.
   up(&spiMutex);
   return 0;
}

static ssize_t avr32io_Read(struct file *filp, char __user *userBuf, size_t len, loff_t *f_pos)
{
   int ret;
   u8 toUser[len];
   int minor = iminor(filp->f_dentry->d_inode);

   DEBUG("avr32io_Read: minor %i isAin: %i isDin: %i\n", minor, IS_AIN(minor), IS_DIN(minor));

   if(IS_AIN(minor)){ //Analog In
      if(len != 2) return -EPERM;
      ret = Cmd_GetAIn(toUser, MinorToChAin(minor));
      if(ret < 0)return ret;
      copy_to_user(userBuf, toUser, len);
      return len;
}

   if(IS_DIN(minor)){ //Digital In
      if(len != 1) return -EPERM;
      ret = Cmd_GetDIn(toUser);
      if(ret < 0) return ret;
      copy_to_user(userBuf, toUser, len);
      return len;
   }

   //If no output matches the minor number the write operation fails.
   return -EPERM;
}

static ssize_t avr32io_Write(struct file *filp, const char __user *userBuf, size_t len, loff_t *f_pos)
{
   u8 fromUser[len];
   int minor = iminor(filp->f_dentry->d_inode);

   copy_from_user(&fromUser, userBuf, len);

   DEBUG("avr32io_Write: minor %i isAout: %i isDout: %i\n", minor, IS_AOUT(minor), IS_DOUT(minor));

   if(IS_AOUT(minor)){ //Analog Out
      if(len != 3) return -EPERM;
      if(Cmd_SetAOut(fromUser + 1, MinorToChAout(minor), fromUser[0]) >= 0){
         return len;
      }
   }

   if(IS_DOUT(minor)){ //Digital Out
      if(len != 2) return -EPERM;
      if(Cmd_SetDOut(fromUser) >= 0){
         return len;
      }
   }

   //If no output matches the minor number the write operation fails.
   return -EPERM;
}

static int avr32io_Ioctl(struct inode	*node, struct file *filp, unsigned int cmd, unsigned long arg)
{
   int minor = iminor(filp->f_dentry->d_inode);

   DEBUG("avr32io_Ioctl: ioctl minor: %i cmd: %i isAout: %i isAin: %i\n", minor, cmd, IS_AOUT(minor), IS_AIN(minor));

   switch(cmd){
      case AVR32IO_IOCTL_RES_AOUT: // Change resolution on a analog channel.
         if(IS_AOUT(minor)){
            return Cmd_SetAOutResolution((u8)(arg), MinorToChAout(minor));
         }
#ifdef BUFFERING_MODE
      case AVR32IO_IOCTL_EN_AIN: // Enables or disables analog in.
         if(IS_AIN(minor)){
            return Cmd_EnableAIn((u8)(arg), MinorToChAin(minor));
         }
#endif
   }

   // If cmd is not a valid command.
   return -ENOIOCTLCMD;
}

