#ifndef __KM_AVR32_IO_H
#define __KM_AVR32_IO_H

// Includes
#include <linux/module.h>
#include <linux/init.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/device.h>
#include <linux/spi/spi.h>
#include <linux/ioctl.h>
#include <linux/delay.h>
#include <asm/uaccess.h>
#include <asm-avr32/io.h>
#include <linux/types.h>
#include <linux/wait.h>
#include <linux/interrupt.h>

//Defines IOCTL numbers.
#define AVR32IO_IOCTL_MAGIC 'k'
#define AVR32IO_IOCTL_RES_AOUT _IOW(AVR32IO_IOCTL_MAGIC, 0, int)
#define AVR32IO_IOCTL_EN_AIN _IOW(AVR32IO_IOCTL_MAGIC, 1, int)

// Defines if driver should print out debugging information.
//#define VERBOSE_DEBUG
#ifdef VERBOSE_DEBUG
	#define DEBUG(fmt, args...) { if(printk_ratelimit()) printk(KERN_ALERT fmt, ## args); }
#else
	#define DEBUG(fmt, args...)
#endif

// Defines if buffering mode or interrupt mode will be used for ADC.
#define BUFFERING_MODE
//#define INTERRUPT_MODE

// Runs when kernel is started/module is installed. Initialize the needed data structures.
static int __init avr32io_Init(void);

// Runs when module is removed. Removes data structures.
static void __exit avr32io_Exit(void);

// Runs when spi driver is started. Initialize the SPI driver.
static int __devinit avr32io_Probe(struct spi_device * spi);

// Runs when an user-space program opens a device node. Tries to mark the SPI as busy, error if SPI already is busy.
static int avr32io_Open(struct inode *node,struct file *filp);

// Runs when the last user-space program closes a device node. mark the SPI as not busy.
static int avr32io_Release(struct inode *node,struct file *filp);

// Runs when an user-space program tries to read from a device node. Send command to I/O-card and return the result.
static ssize_t avr32io_Read(struct file *filp, char __user *userBuf, size_t len, loff_t *f_pos);

// Runs when an user-space program tries to write to a device node. Send command to I/O-card.
static ssize_t avr32io_Write(struct file *filp, const char __user *userBuf, size_t len, loff_t *f_pos);

// Runs when an user-space program tries to send an IO control command. Send the command to I/O-card.
static int avr32io_Ioctl(struct inode	*node, struct file *filp, unsigned int cmd, unsigned long arg);

// Struct containing data structures for the kernel module.
struct avr32io_device {
   dev_t number; // Major and minor numbers.
   struct cdev *charDevice; // Char device.
   struct spi_device *spi; // SPI device.
   int busy; // Device busy.
#ifdef INTERRUPT_MODE
   int irqID; // Interrupt used for analog in.
   int irqFlag; // Flag when receiving interrupt.
#endif
};

// Pointer to device used in driver.
static struct avr32io_device * device;

// Struct that defines the functions used for different file operations.
static struct file_operations fops = {
   .owner = THIS_MODULE,
   .read = avr32io_Read,
   .write = avr32io_Write,
   .open = avr32io_Open,
   .release = avr32io_Release,
   .ioctl = avr32io_Ioctl,
};

// Struct that defines the SPI driver.
static struct spi_driver avr32io_driver = {
   .probe = avr32io_Probe,
   .driver = {
      .name = "avr32io",
   }
};

DECLARE_MUTEX(spiMutex);

// Includes.
#include "../minorNumbers.h"
#include "../spiByteCodes.h"
#include "avr32io_SPI.h"
#include "avr32io_Cmd.h"

// Defines init and exit functions.
module_init(avr32io_Init);
module_exit(avr32io_Exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Oyvind Netland, NTNU, Trondheim, Norway");
MODULE_DESCRIPTION("Device driver for AVR32 I/O-card developed by Oyvind Netland, during a master thesis on NTNU, Trondheim, Norway");

#endif // __KM_AVR32_IO_H
